;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Copyright   : (c)1998 CMU/Mike Byrne, All Rights Reserved
;;; Availability: public domain
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : byrne+@andrew.cmu.edu
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : virtual-test-suite.lisp
;;; Version     : 2.0b5
;;; 
;;; Description : Lisp source for the "test suite" example with ACT-R/PM, but
;;;             : with a "virtual" window rather than an actual window. 
;;;             : Changes from the "normal" MCL window stuff are commented.
;;;             : 
;;;             : Most of this is pretty straightforward but it does give
;;;             : examples of how to use some of the ACT-R/PM stuff like sound
;;;             : events (both generating and responding to), the device hook
;;;             : function, handling mouse and keyboard hits, tracking, and
;;;             : the virtual-view interface with the Vision Module.
;;; 
;;; ----- History -----
;;; 98.06.05 Mike Byrne
;;;             : Documentation genesis.
;;; 98.08.25 mdb
;;;             : Added AFTER method for OUTPUT-CURSOR-MOVE.
;;; 99.04.07 mdb
;;;             : Updated for use with virtual-view stuff.
;;; 00.09.13 mdb
;;;             : Updated for change to vectors.
;;; 00.11.15 mdb
;;;             : Minor fix in DEVICE-HANDLE-CLICK.
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defvar *experiment* nil "The window")
(defparameter *letters '("a" "b" "c" "d" "e" "f" "g" "h" "i" "j" "k" "l" "m"))

(defun flip ()
  (= 0 (random 2)))

(defclass test-window (virtual-window)          ; delta for virtual
  ((base-states :accessor base-states
                :initform '(:start :sound :key1 :key2 :key3 :click))
   (states :accessor states)
   (current-state :accessor current-state)
   (start-time :accessor start-time :initform 0)
   (first-loc :accessor first-loc :initform nil)
   (move1 :accessor move1 :initform nil)
   (move2 :accessor move2 :initform nil)
   (last-update :accessor last-update :initform 0)
   )
  (:default-initargs
    :width 300                          ; delta for virtual
    :height 300))                       ; delta for virtual


(defmethod initialize-instance :after ((tw test-window) &key)
  (setf (states tw) (copy-list (base-states tw)))
  (setf (current-state tw) (pop (states tw)))
  (make-start tw)
  (setf (start-time tw) 0.0))


(defmethod make-start ((tw test-window))
  "Make the 'start' button at a random location."
  (let ((x (+ 50 (random 200)))
        (y (+ 50 (random 200))))
    (add-subviews tw
      (make-instance 'button-vdi        ; all this is delta for virtual
        :x-pos x :y-pos y
        :dialog-item-text "Start"
        :action #'(lambda (self)
                            (start-btn (view-window self)))))))

#| delta for virtual
      (make-dialog-item 'button-dialog-item
                        (make-point x y)
                        #@(60 20)
                        "Start"
                        #'(lambda (self)
                            (start-btn (view-window self)))
                        :default-button nil))))
|#


(defmethod clear-window ((tw test-window))
  "Remove all subviews and have RPM process the screen."
  (apply #'remove-subviews tw (subviews tw))
  (pm-proc-display :clear t))


(defmethod print-rt ((tw test-window))
  "Prints the RT, pops the state, and clears the window."
  (format t "~&[-=- DEVICE -=-] RT was: ~S" (- (pm-get-time) (start-time tw)))
  (setf (current-state tw) (pop (states tw)))
  (clear-window tw))


;;; START-BTN      [Method]
;;; Description : When the "Start" button is pressed, there are a few
;;;             : things to do.  Print the RT (which also does some 
;;;             : other bookkeeping), then picks a random onset for
;;;             : the audio stimulus, and randomly picks which frequency
;;;             : to use.

(defmethod start-btn ((tw test-window))
  "Response when the 'start' button is pressed."
  (print-rt tw)
  (let ((onset (float (/ (+ 150 (random 300)) 1000)))
        (freq (if (flip) 800 2000)))
    (new-tone-sound freq onset)
    (pm-delayed-event onset #'beep)
    (setf (start-time tw) (+ (pm-get-time) (* 1000 onset)))))


(defmethod got-sound ((tw test-window) text)
  "Handle the sound.  Print the RT and what was said, and show the letter."
  (print-rt tw)
  (format t "~&[-=- DEVICE -=-] Word was: ~S" text)
  (setf (first-loc tw) (show-text tw (random-item *letters))))


(defmethod show-text ((tw test-window) (str string) &optional x y)
  "Show text at some random location, draw it, reset the clock, and process the screen."
  (let ((x (if x x (+ 50 (random 200))))
        (y (if y y (+ 50 (random 200)))))
    (add-subviews tw
      ;; a bunch of deltas for virtual
      (make-instance 'static-text-vdi :x-pos x :y-pos y 
                        :dialog-item-text str))
    ;(view-draw-contents tw)
    ;(event-dispatch)
    (setf (start-time tw) (pm-get-time))
    (pm-proc-display)
    (vector x y)))


;;; SETUP-TRACKING      [Method]
;;; Description : When we start the tracking business, we need to first 
;;;             : set up the star and oval views--randomly decide which
;;;             : is the left view and which is the right.  Then set the
;;;             : locations randomly.  Add them and process the screen.
;;;             : Finally, set the updater function and the start time.

(defmethod setup-tracking ((tw test-window))
  "Set up tracking:  Install the views, set update information."
  ;; install the two views
  (cond ((flip) 
         (setf (move1 tw) (make-instance 'circle-view))
         (setf (move2 tw) (make-instance 'star-view)))
        (t
         (setf (move2 tw) (make-instance 'circle-view))
         (setf (move1 tw) (make-instance 'star-view))))
  (set-view-position (move1 tw)
                     (random 120) (+ 150 (random 120)))
  (set-view-position (move2 tw)
                     (+ 150 (random 120)) (+ 150 (random 120)))
  (add-subviews tw (move1 tw) (move2 tw))
  (setf (current-state tw) :TRACKING)
  (pm-proc-display)
  (setf (last-update tw) (mp-time *mp*)))


(defmethod move-views ((tw test-window))
  "Change the location of the moving views, update the screen, and process the screen."
  (let ((x1 (px (view-position (move1 tw))))
        (y1 (py (view-position (move1 tw))))
        (x2 (px (view-position (move2 tw))))
        (y2 (py (view-position (move2 tw)))))
    (incf x1 (random 10))
    (decf y1 (random 10))
    (decf x2 (random 10))
    (decf y2 (random 10))
    (when (< 270 x1) (setf x1 270))
    (when (> 0 y1) (setf y1 0))
    (when (> 0 x2) (setf x2 0))
    (when (> 0 y2) (setf y2 0))
    (set-view-position (move1 tw) x1 y1)
    (set-view-position (move2 tw) x2 y2)
    ;(event-dispatch)                    ; not for virtual window
    (pm-proc-display)))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Being an RPM device.
;;;; ---------------------------------------------------------------------- ;;;;


;;; DEVICE-HANDLE-KEYPRESS      [Method]
;;; Description : What is done when a key is pressed depends on the state 
;;;             : we're currently in.  Either show the appropriate text
;;;             : or start the tracking business.

(defmethod device-handle-keypress ((tw test-window) key)
  "Handle keystrokes."
  (print-rt tw)
  (format t "~&[-=- DEVICE -=-] Got key: ~A" key)
  (ecase (current-state tw)
    (:key2 (show-text tw (random-item '("up" "down"))))
    (:key3 (show-text tw 
                      (random-item '("key left" "key right"))
                      (px (first-loc tw)) (py (first-loc tw))))
    (:click (setup-tracking tw))))


;;; DEVICE-HANDLE-CLICK      [Method]
;;; Description : When the view is clicked, what to do depends on the 
;;;             : state.  If we're waiting for the "Start" button to be pressed, 
;;;             : we have to pass the event so the button will get it.
;;;             : A mouse click can also be used in place of the sound in cases
;;;             : where the sound won't be listened for.
;;;             : Finally, the real click we're handling here is the one
;;;             : that ends tracking.

(defmethod device-handle-click ((tw test-window))
  "Handle a click on the view."
  (when (eq (current-state tw) :start) 
    (call-next-method)
    (return-from device-handle-click nil))
  (if (eq (current-state tw) :sound)
    (got-sound tw "click")
    (progn  
      (print-rt tw)
      (pm-set-params :device-updater-fct #'null))))
  

(defmethod device-update ((tw test-window) time)
  (when (eq (current-state tw) :TRACKING)
    (when (<= 0.050 (- time (last-update tw)))
      (setf (last-update tw) time)
      (move-views tw))))  


(defmethod device-speak-string ((tw test-window) string)
  (got-sound tw string))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; The moving subviews
;;;; ---------------------------------------------------------------------- ;;;;

;;; CIRCLE-VIEW      [Class]
;;; Description : A class for the moving circle on the screen--nothing special
;;;             : is needed in the class itself.

(defclass circle-view (virtual-view)    ;  delta for virtual
  ()
  (:default-initargs
    :height 30                          ;  delta for virtual
    :width 30))                         ;  delta for virtual

#| delta for virtual
;;; VIEW-DRAW-CONTENTS      [Method]
;;; Description : Drawing the view on the screen is just a FILL-OVAL on
;;;             : the view.


(defmethod view-draw-contents ((self circle-view))
  (fill-oval self *black-pattern* 0 0 (point-h (view-size self)) 
             (point-v (view-size self))))
|#


;;; BUILD-FEATURES-FOR      [Method]
;;; Description : How to turn the view into a feature for the icon.  
;;;             : OVAL-FEATURE works as the class because that's a feature
;;;             : that's already defined in RPM.

(defmethod build-features-for ((self circle-view) (vis-mgr vision-module))
  (make-instance 'oval-feature
    :x (px (view-loc self))
    :y (py (view-loc self))
    :screen-obj self
    :width (width self)
    :height (height self)))



;;; STAR-FEATURE      [Class]
;;; Description : The ICON-FEATURE class for the star.  If we had any other
;;;             : information that had to be in the feature, we would also
;;;             : need a FEAT-TO-DMO method to translate the feature to 
;;;             : a declarative memory representation.  Not necessary since
;;;             : the default is to build a chunk with the type and value
;;;             : specified.  (Do need a "STAR" chunk-type, though.)

(defclass star-feature (icon-feature)
  ()
  (:default-initargs
    :kind 'STAR
    :value 'STAR
    :dmo-id (gentemp "STAR")))


;;; STAR-VIEW      [Class]
;;; Description : A class for the star-view itself.  No slots, there's no
;;;             : special information to be stored here.

(defclass star-view (virtual-view)      ; delta for virtual
  ()
  (:default-initargs
    :height 30                          ; delta for virtual
    :width 30))                         ; delta for virtual


#| delta for virtual
;;; VIEW-DRAW-CONTENTS      [Method]
;;; Description : The drawing method for the star view.  Draw the four lines,
;;;             : two corner-to-corner and two midpoint-to-midpoint.

(defmethod view-draw-contents ((self star-view))
  (let ((xmax (point-h (view-size self)))
        (ymax (point-v (view-size self))))
    (move-to self 0 0)
    (line-to self xmax ymax)
    (move-to self xmax 0)
    (line-to self 0 ymax)
    (move-to self 0 (/ ymax 2))
    (line-to self xmax (/ ymax 2))
    (move-to self (/ xmax 2) 0)
    (line-to self (/ xmax 2) ymax)))
|#


;;; BUILD-FEATURES-FOR      [Method]
;;; Description : Creates an icon feature for the star view, which is
;;;             : one instance of the star-feature.

(defmethod build-features-for ((self star-view) (vis-mgr vision-module))
  (make-instance 'star-feature
    :x (px (view-loc self))
    :y (py (view-loc self))
    :screen-obj self
    :width (width self)
    :height (height self)))
