;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Copyright   : (c)2001-2003 Rice U./Mike Byrne, All Rights Reserved
;;; Availability: public domain
;;; Address     : Rice University
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : cognition-module.lisp
;;; Version     : 2.1b5
;;; 
;;; Description : Class for wrapping around ACT-R to make cognition a
;;;             : proper object for the scheduler
;;; 
;;; Bugs        : 
;;; 
;;; Todo        : [] :AFTER methods to handle buffer stuffing
;;; 
;;; ----- History -----
;;; 01.07.25 Mike Byrne
;;;             :  Incept date.
;;; 01.08.07 mdb
;;;             : Added UPDATE-MODULE and UPDATE-DM-STATE dummy methods.
;;; 01.11.28 mdb/Dan Bothell [b4]
;;;             : Changed EXECUTE-RHS and COMPLETE-RETRIEVAL to deal with
;;;             : time issues.
;;; 02.02.14 Dan [b5]
;;;             : added two slots to the cognitive-module and modified teh
;;;             : select production method because of the way ACT-R uses 
;;;             : global variables for it's parameters - matching sets them 
;;;             : then action uses them, but matching needs the original so it 
;;;             : has to be stored somewhere before a run.  Also see actr-interface
;;;             : for the updated pm-run to handle this.  Like I said
;;;             : there, a better solution is probably the right answer...
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *rand-ps* nil)

(defclass cognition-module (pm-module)
  ((ps-state :accessor ps-state :initarg :ps-state :initform :idle)
   (last-retrieval-request :accessor last-retrieval-request 
                           :initarg :last-retrieval-request :initform nil)
   (min-run-time :accessor min-run-time :initarg :min-run-time 
                 :initform 0.050)
   (rt :accessor rt :initform nil)
   (ut :accessor ut :initform 0.0)
   )
  (:default-initargs
    :version-string "5.0.6"
    :name :COGNITION
    ))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; hadling standard RPM calls

(defmethod reset-module :after ((cog-mod cognition-module))
  (setf (ps-state cog-mod) :IDLE)
  (setf (last-retrieval-request cog-mod) nil))


;;; RUN-MODULE      [Method]
;;; Date        : 01.07.27
;;; Description : Has a slightly different RUN-MODULE method here because we
;;;             : don't want trace output (this is handled by ACT-R 5) and
;;;             : don't need to modify state.

(defmethod run-module ((module cognition-module) the-time)
  (setf (state-change module) nil)
  (while (and (input-q module)
              (>= (ms-round the-time)
                  (ms-round (time-tag (first (input-q module))))))
    (let* ((input-entry (pop (input-q module)))
           (the-cmd (first (params input-entry))))
      (when (functionp (step-hook *mp*))
        (funcall (step-hook *mp*) input-entry))
      (apply the-cmd (append (list module) (rest (params input-entry)))))))


(defmethod update-module ((cog-mod cognition-module))
  nil)


(defmethod update-dm-state ((cog-mod cognition-module))
  nil)

(defmethod print-module-state ((cog-mod cognition-module))
  (format t "~& ACT-R's state is ~A" (ps-state cog-mod)))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Module messages for interacting with ACT-R

(defgeneric select-production (cog-mod &optional max-time)
  (:documentation "Request that the production system select a production, given a maximum time constraint."))


;;; SELECT-PRODUCTION      [Method]
;;; Description : A number of things have to be done here.  It's necessary to
;;;             : first check the goal stack, then actually call ACT-R5's
;;;             : MATCHING-CYCLE function, then deal with that result and
;;;             : either schedule the completion of the cycle or note the
;;;             : failure.

(defmethod select-production ((cog-mod cognition-module) 
                                 &optional max-time)
  (if (null (first (no-output (goal-stack-fct))))
    (setf (ps-state cog-mod) :NO-GOAL)
    (let ((action-latency 
           (matching-cycle max-time (rt cog-mod) (ut cog-mod))))
      
      (if action-latency
        (progn
          (when *rand-ps* (setf action-latency (rand-time action-latency)))
          (setf (ps-state cog-mod) :PENDING-EXECUTION)
	  ;; what about time randomization?
          (queue-command :time action-latency 
                         :where :COGNITION 
                         :command 'execute-rhs 
                         :params (list action-latency)
                         :randomize nil
                         ))
        (setf (ps-state cog-mod) :FAILED-SELECTION)))))


;;; EXECUTE-RHS      [Method]
;;; Date        : 01.07.27
;;; Description : Calls the ACT-R5 function ACTION-CYCLE, idles the CP [which
;;;             : will cause a new SELECT-PRODUCTION to be scheduled by the
;;;             : Master Process], and then it's necessary to deal with the
;;;             : retrieval, if any.

(defmethod execute-rhs ((cog-mod cognition-module) (latency number))
  (setf *time* (- *time* latency)) ;; back up in time because action-cycle is
  ;; going to increment *time* and I can't just
  ;; pass 0 because it needs that latency for
  ;; parameter learning purposes...
  (action-cycle latency)
  (setf (ps-state cog-mod) :IDLE)
  (when (and *retrieval-scheduler* 
             (not (equal *retrieval-scheduler* 
                         (last-retrieval-request cog-mod))))
    (setf (last-retrieval-request cog-mod) (copy-tree *retrieval-scheduler*))
    ;; have to subtract because the first of *retrieval-schuduler* is the
    ;; absolute time of the retrieval completion
    (queue-command :time (- (first *retrieval-scheduler*) (mp-time *mp*))
                   :where :COGNITION
                   :command 'complete-retrieval
                   :params (list (first *retrieval-scheduler*))
                   :randomize nil)))



;;; don't really need the time parameter--the MP knows what time it is.
(defmethod complete-retrieval ((cog-mod cognition-module) (time number))
  (when *retrieval-scheduler*
    (retrieval-event time)))


