;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Copyright   : (c)1998-2003 CMU/Rice U./Mike Byrne, All Rights Reserved
;;; Availability: public domain
;;; Address     : Rice University
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : dmi.lisp
;;; Version     : 2.1.1
;;; 
;;; Description : The "Declarative Memory Interface"
;;;             : Defines a class ("declarative memory object," or DMO)
;;;             : and numerous methods to provide an interface layer
;;;             : between the PM stuff and the specific production
;;;             : system in terms of referring to declarative
;;;             : memory elements.  Cleans up a bunch of the PM code.
;;; 
;;; Bugs        : 
;;; 
;;; Todo        : 
;;; 
;;; ----- History -----
;;; 01.07.27 mdb
;;;             : Removed CLEAR-LOC for 2.1/ACT5
;;; 2003.01.21 mdb
;;;             : Removed PS-PTR slot and made it a method instead. 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass dmo ()
  (;(ps-ptr :accessor ps-ptr :initform nil :initarg :ps-ptr)
   (id :accessor id :initform (new-name-fct "DMO") :initarg :id)
   (dmtype :accessor dmtype :initarg :dmtype :initform nil)
   (pm-obj :accessor pm-obj :initform nil :initarg :pm-obj)))


;;; generic methods 

(defmethod print-object ((self dmo) stream)
  (print-unreadable-object (self stream :type t)
    (princ (id self) stream)))

(defgeneric ps-ptr (self)
  (:documentation "Return the production system [hence PS] declarative memory."))



(defgeneric find-loc-dmo (loc)
  (:documentation "Given an XY location, return the DMO representing that location, if present."))

(defmethod find-loc-dmo ((loc vector))
  (random-item 
   (matching-dmos 'visual-location
                  `(screen-x ,(px loc) screen-y ,(py loc)))))

(defmethod find-loc-dmo ((loc list))
  (find-loc-dmo (coerce loc 'vector)))


(defgeneric xy-to-dmo (loc attended-p)
  (:documentation "Given an XY loc and an attended value, find the declarative memory representaion for that location, or build it.  Either way, set the attended flag of the DMO."))

(defmethod xy-to-dmo ((loc vector) (attended-p symbol))
  (let ((loc-dmo (find-loc-dmo loc))
        (name (new-name-fct "LOC")))
    (if loc-dmo
      (if attended-p 
        (set-attributes loc-dmo '(attended t))
        loc-dmo)
      (make-dme name 'visual-location  
                `(screen-x ,(px loc) screen-y ,(py loc)
                           attended ,attended-p objects nil)
                :where :external))))


(defmethod xy-to-dmo ((loc list) (attended-p symbol))
  (xy-to-dmo (coerce loc 'vector) attended-p))


#|
(defgeneric clear-loc (dmo)
  (:documentation "Clears the <objects> slot of a DMO."))

(defmethod clear-loc ((self dmo))
  (set-attributes self '(objects nil)))
|#

(defgeneric dmo-to-xy (dmo)
  (:documentation "Takes a location DMO and returns the XY location."))

(defmethod dmo-to-xy ((self dmo))
  (vector (get-attribute self 'screen-x) (get-attribute self 'screen-y)))



;;; Must be overridden with PS-specific methods.

(defgeneric make-dme (id dmtype attrs &key obj where)
  (:documentation "Creates both a PS-specific DM rep and an abstract DMO for a specification."))

(defmethod make-dme (id (dmtype symbol) (attrs list) &key (obj nil) (where nil))
  (declare (ignore id dmtype attrs obj where))
  (error "No method defined for MAKE-DME."))


(defgeneric get-attribute (dmo attrname)
  (:documentation "Retuns the value of a DMO's specified attribute."))

(defmethod get-attribute ((self dmo) (attrname symbol))
  (declare (ignore self attrname))
  (error "No method defined for GET-ATTRIBUTE."))


(defgeneric set-attributes (dmo attrs)
  (:documentation "Sets the value of one or more attributes of a DMO."))

(defmethod set-attributes ((self dmo) (attrs list))
  (declare (ignore self attrs))
  (error "No method defined for SET-ATTRIBUTES."))


(defgeneric psdme-to-dmo (psdme)
  (:documentation "Given a PS-specific DME, return the corresponding DMO."))

(defmethod psdme-to-dmo ((psdme null))
  nil)

(defmethod psdme-to-dmo (psdme)
  (declare (ignore psdme))
  (error "No method defined for PSDME-TO-DMO."))


(defgeneric dmo-to-psdme (dmo)
  (:documentation "Return the PS-specific DME form for a given DMO."))

(defmethod dmo-to-psdme ((self null))
  nil)

(defmethod dmo-to-psdme ((self dmo))
  (declare (ignore self))
  (error "No method defined for DMO-TO-PSDME."))


(defgeneric matching-dmos (dmtype attrs)
  (:documentation "Return a list of DMOs matching the given spec."))

(defmethod matching-dmos ((dmtype symbol) (attrs list))
  (declare (ignore self))
  (error "No method defined for DMO-TO-PSDME."))
