;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Copyright   : (c)2000 Rice U./Mike Byrne, All Rights Reserved
;;; Availability: public domain
;;; Address     : Rice University
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : virtual-cursor.lisp
;;; Version     : r2
;;; 
;;; Description : Provides a view-based virtual cursor for RPM.
;;; 
;;; Bugs        : The central problem is getting clicks to work.  The "easy"
;;;             : thing to do is just move the real cursor when a click is
;;;             : requested, except that doesn't seem to work if the real 
;;;             : cursor is in motion at the time.
;;; 
;;; Todo        : 
;;; 
;;; ----- History -----
;;; 00.08.07 Mike Byrne
;;;             :  Incept date.
;;; 01.07.02 mdb
;;;             : Fixed CURRENT-MOUSE call to be CURRENT-CURSOR.
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *virtual-cursor* nil "The virtual cursor.")




(defclass virtual-cursor (view)
  ()
  (:default-initargs
       :view-size #@(10 10)
       ))


;;; VIEW-DRAW-CONTENTS      [Method]
;;; Description :  Draws a green triangle.

(defmethod view-draw-contents ((self virtual-cursor))
  (let ((oldmode (pen-mode self))
        (oldpat (pen-pattern self))
        (oldsize (pen-size self)))
    (set-pen-mode self :pator) 
    (set-pen-pattern self *black-pattern*)
    (set-pen-size self 3 3)
    (with-focused-view self
      (with-fore-color *green-color*
        (move-to self 0 0)
        (line-to self 8 3)
        (line-to self 3 8)
        (line-to self 0 0)
        ;(line-to self 10 10)
        ))
    (set-pen-mode self oldmode)
    (set-pen-pattern self oldpat)
    (set-pen-size self (point-h oldsize) (point-v oldsize))
    ))

;;; As with the focus ring, want to be sure we don't trap clicks or 
;;; or generate features.

(defmethod build-features-for ((self virtual-cursor) (vis-mod vision-module))
  nil)

(defmethod point-in-click-region-p ((self virtual-cursor) where)
  (declare (ignore where))
  nil)


(defmethod cursor-to-feature ((wind window))
  (when (equal (view-window *virtual-cursor*) wind)
    (let ((loc (current-cursor (motor-m *mp*))))
      (make-instance 'cursor-feature
        :x (px loc) :y (py loc) :value 'POINTER))))


;;; device-level stuff


;;; DEVICE-MOVE-CURSOR-TO      [Method]
;;; Description : Moving the cursor is just moving the view and redrawing.
;;;             : The only problem is that the whole window has to redraw,
;;;             : rather than just the cursor, because asking just the 
;;;             : cursor to redraw often bombs.

(defmethod device-move-cursor-to ((device window) (xyloc vector))
  (set-view-position *virtual-cursor* (px xyloc) (py xyloc))
  (add-subviews device *virtual-cursor*)
  (event-dispatch)
  (view-draw-contents device)           ; *virtual-cursor* doesn't work.
  (event-dispatch))



(defmethod device-move-cursor-to ((device window) (xyloc list))
  (device-move-cursor-to device (coerce xyloc 'vector)))


(defmethod get-mouse-coordinates ((device window))
  (p2vpt (view-position *virtual-cursor*)))


;;; DEVICE-HANDLE-CLICK      [Method]
;;; Description : Store the location of the real mouse, then move to the click
;;;             : point, then click, then move the mouse back.  Simply 
;;;             : calling VIEW-CLICK-EVENT-HANDLER directly should work, but
;;;             : doesn't because somewhere in the hierarchy, #_GetMouse 
;;;             : gets called and the click will be missed.  
;;;
;;;             : The other issue is that this doesn't seem to work right
;;;             : when the cursor is in motion when this is callled.  I
;;;             : have no idea why.

(defmethod device-handle-click ((device window))
  ;; save the old mouse position of the real mouse
  (let ((old-mouse-pt (view-mouse-position device))
        (new-pt (view-position *virtual-cursor*))
        )
    ;; move the real mouse to here
    (while (not (= (view-mouse-position device) new-pt))
      (fast-move-cursor-to device new-pt)
      ;(event-dispatch)
      ;(print (point-string (view-mouse-position device)))
      )
    ;; do the click
    (view-click-event-handler device new-pt)
    (event-dispatch)
    ;; restore the real mouse
    (fast-move-cursor-to device old-mouse-pt)
    (event-dispatch)
    ))


(defmethod fast-move-cursor-to ((wind window) (new-pt number))
  (let ((absloc (local-to-global wind new-pt)))
    (without-interrupts
     (ccl::%put-point (%int-to-ptr #$MTemp) absloc)
     (ccl::%put-point (%int-to-ptr #$RawMouse) absloc)
     (%put-word (%int-to-ptr #$CrsrNew) -1))
    (while (eql (%get-signed-word (%int-to-ptr #$CrsrNew)) -1))))



;;; MOVE-CURSOR-TO      [Method]
;;; Description : Poke the cursor around, then spin until the low memory 
;;;             : knows it's there, then update the cursor, then spin again
;;;             : until

(when (fboundp 'move-cursor-to) (unintern 'move-cursor-to))


(defmethod move-cursor-to ((wind window) (new-pt number))
  (let ((absloc (local-to-global wind new-pt)))
    (without-interrupts
     (ccl::%put-point (%int-to-ptr #$MTemp) absloc)
     (ccl::%put-point (%int-to-ptr #$RawMouse) absloc)
     (%put-word (%int-to-ptr #$CrsrNew) -1))
    (while (eql (%get-signed-word (%int-to-ptr #$CrsrNew)) -1))
    (update-cursor)
    (while (not (= new-pt (view-mouse-position wind)))
      (event-dispatch))))


(defmethod move-cursor-to ((wind window) (xyloc list))
  (move-cursor-to wind (xy->point xyloc)))

(defmethod move-cursor-to ((wind window) (xyloc vector))
  (move-cursor-to wind (vpt2p xyloc)))



(defmethod device-update :after ((device window) time)
  (declare (ignore time))
  (pm-set-cursor-position-fct (current-cursor (motor-m *mp*))))


(eval-when (load eval)
  (setf *virtual-cursor* (make-instance 'virtual-cursor)))